/*
 *  linux/drivers/mmc/host/wmt-sdmmc.c - VT8500/WM8505 AHB to MMC/SD driver
 *
 *  Copyright (C) 2010 Tony Prisk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/dma-mapping.h>
#include <linux/highmem.h>
#include <linux/delay.h>

#include <linux/mmc/host.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/sd.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/byteorder.h>


#define DRIVER_NAME "wmt-sdmmc"


/* MMC/SD controller registers (VT8500/WM8505) */
#define SDMMC_CTLR			0x00
#define SDMMC_CMD			0x01
#define SDMMC_RSPTYPE			0x02
#define SDMMC_ARG			0x04
#define SDMMC_BUSMODE			0x08
#define SDMMC_BLKLEN			0x0C
#define SDMMC_BLKCNT			0x0E
#define SDMMC_RSP			0x10
#define SDMMC_CBCR			0x20
#define SDMMC_INTMASK0			0x24
#define SDMMC_INTMASK1			0x25
#define SDMMC_STS0			0x28
#define SDMMC_STS1			0x29
#define SDMMC_STS2			0x2A
#define SDMMC_STS3			0x2B
#define SDMMC_RSPTIMEOUT		0x2C
#define SDMMC_CLK			0x30	/* VT8500 only */
#define SDMMC_EXTCTRL			0x34
#define SDMMC_SBLKLEN			0x38
#define SDMMC_DMATIMEOUT		0x3C


/* SDMMC_CTLR bit fields */
#define CTLR_CMD_START			0x01
#define CTLR_CMD_WRITE			0x04
#define CTLR_FIFO_RESET			0x08

/* SDMMC_BUSMODE bit fields */
#define BM_SPI_MODE			0x01
#define BM_FOURBIT_MODE			0x02
#define BM_EIGHTBIT_MODE		0x04
#define BM_SD_OFF			0x10
#define BM_SPI_CS			0x20
#define BM_SD_POWER			0x40
#define BM_SOFT_RESET			0x80
#define BM_ONEBIT_MASK			0xFD

/* SDMMC_BLKLEN bit fields */
#define BLKL_CRCERR_ABORT		0x0800
#define BLKL_CD_POL_HIGH		0x1000
#define BLKL_GPI_CD			0x2000
#define BLKL_DATA3_CD			0x4000
#define BLKL_INT_ENABLE			0x8000

/* SDMMC_INTMASK0 bit fields */
#define INT0_MBLK_TRAN_DONE_INT_EN	0x10
#define INT0_BLK_TRAN_DONE_INT_EN	0x20
#define INT0_CD_INT_EN			0x40
#define INT0_DI_INT_EN			0x80

/* SDMMC_INTMASK1 bit fields */
#define INT1_CMD_RES_TRAN_DONE_INT_EN	0x02
#define INT1_CMD_RES_ROUT_INT_EN	0x04
#define INT1_MBLK_AUTO_STOP_INT_EN	0x08
#define INT1_DATA_TOUT_INT_EN		0x10
#define INT1_RESCRC_ERR_INT_EN		0x20
#define INT1_RCRC_ERR_INT_EN		0x40
#define INT1_WCRC_ERR_INT_EN		0x80

/* SDMMC_STS0 bit fields */
#define STS0_WRITE_PROTECT		0x02
#define STS0_CD_DATA3			0x04
#define STS0_CD_GPI			0x08
#define STS0_MBLK_DONE			0x10
#define STS0_BLK_DONE			0x20
#define STS0_CARD_DETECT		0x40
#define STS0_DEVICE_INS			0x80

/* SDMMC_STS1 bit fields */
#define STS1_SDIO_INT			0x01
#define STS1_CMDRSP_DONE		0x02
#define STS1_RSP_TIMEOUT		0x04
#define STS1_AUTOSTOP_DONE		0x08
#define STS1_DATA_TIMEOUT		0x10
#define STS1_RSP_CRC_ERR		0x20
#define STS1_RCRC_ERR			0x40
#define STS1_WCRC_ERR			0x80

/* SDMMC_STS2 bit fields */
#define STS2_CMD_RES_BUSY		0x10
#define STS2_DATARSP_BUSY		0x20
#define STS2_DIS_FORCECLK		0x80


/* MMC/SD DMA Controller Registers */
#define SDDMA_GCR			0x100
#define SDDMA_IER			0x104
#define SDDMA_ISR			0x108
#define SDDMA_DESPR			0x10C
#define SDDMA_RBR			0x110
#define SDDMA_DAR			0x114
#define SDDMA_BAR			0x118
#define SDDMA_CPR			0x11C
#define SDDMA_CCR			0x120


/* SDDMA_GCR bit fields */
#define DMA_GCR_DMA_EN			0x00000001
#define DMA_GCR_SOFT_RESET		0x00000100

/* SDDMA_IER bit fields */
#define DMA_IER_INT_EN			0x00000001

/* SDDMA_ISR bit fields */
#define DMA_ISR_INT_STS			0x00000001

/* SDDMA_RBR bit fields */
#define DMA_RBR_FORMAT			0x40000000
#define DMA_RBR_END			0x80000000

/* SDDMA_CCR bit fields */
#define DMA_CCR_RUN			0x00000080
#define DMA_CCR_IF_TO_PERIPHERAL	0x00000000
#define DMA_CCR_PERIPHERAL_TO_IF	0x00400000

/* SDDMA_CCR event status */
#define DMA_CCR_EVT_NO_STATUS		0x00000000
#define DMA_CCR_EVT_UNDERRUN		0x00000001
#define DMA_CCR_EVT_OVERRUN		0x00000002
#define DMA_CCR_EVT_DESP_READ		0x00000003
#define DMA_CCR_EVT_DATA_RW		0x00000004
#define DMA_CCR_EVT_EARLY_END		0x00000005
#define DMA_CCR_EVT_SUCCESS		0x0000000F

struct wmt_dma_descriptor {
	u32 flags;
	u32 data_buffer_addr;
	u32 branch_addr;
	u32 reserved1;
};

struct wmt_mci_priv {
	struct mmc_host *mmc;
	void __iomem *sdmmc_base;

	int irq_regular;
	int irq_dma;

	/* dma descriptor buffer */
	void *dma_desc_buffer;
	dma_addr_t dma_desc_device_addr;

	struct completion *comp_dma;

	/* sd data */
	struct mmc_request *req;
};

static irqreturn_t wmt_mci_dma_isr(int irq_num, void *data)
{
	struct mmc_host *mmc;
	struct wmt_mci_priv *priv;

	int cnt;
	int status;

	priv = (struct wmt_mci_priv *)data;
	mmc = priv->mmc;

	status = readl(priv->sdmmc_base + SDDMA_CCR) & 0x0F;

	if (status != DMA_CCR_EVT_SUCCESS) {
		printk(KERN_ERR "[MMC/SD] DMA Status = %d\n", status);
		priv->req->data->error = -ETIMEDOUT;
		complete(priv->comp_dma);
		return IRQ_HANDLED;
	}

	priv->req->data->error = 0;
	/*
	 * If we read data from the card, wait for the controller to indicate it has finished sending
	 */
	cnt = 0x1000000;
	if (!(readl(priv->sdmmc_base + SDDMA_CCR) & DMA_CCR_PERIPHERAL_TO_IF)) {
		while (!(readl(priv->sdmmc_base + SDMMC_STS0) & (STS0_BLK_DONE | STS0_MBLK_DONE))) {
			cnt--;
			if (cnt == 0) {
				printk(KERN_ERR "[MMC/SD] Control Data Timeout\n");
				priv->req->data->error = -ETIMEDOUT;
			}
		}
	}
	/* disable the SD DMA controller */
	writel(DMA_ISR_INT_STS, priv->sdmmc_base + SDDMA_ISR);
	writel(0, priv->sdmmc_base + SDDMA_IER);

	complete(priv->comp_dma);

	return IRQ_HANDLED;
}

static irqreturn_t wmt_mci_regular_isr(int irq_num, void *data)
{
	struct wmt_mci_priv *priv;
	u32 status0;
	u32 status1;
	u32 status2;
	u32 reg_tmp;

	priv = (struct wmt_mci_priv *)data;

	status0 = readb(priv->sdmmc_base + SDMMC_STS0);
	status1 = readb(priv->sdmmc_base + SDMMC_STS1);
	status2 = readb(priv->sdmmc_base + SDMMC_STS2);

	/* Check for card insertion */	
	reg_tmp = readb(priv->sdmmc_base + SDMMC_INTMASK0);
	if ( (reg_tmp & INT0_DI_INT_EN) && (status0 & STS0_DEVICE_INS) ) {
		mmc_detect_change(priv->mmc, 0);
		writeb(STS0_DEVICE_INS, priv->sdmmc_base + SDMMC_STS0);
		return IRQ_HANDLED;
	}

	return IRQ_HANDLED;
}

/*
 *  This code is hacky and should be fixed
 *  Hardcoded GPIO bases should be passed from board definition
 *  This code enables the GPIO pin that power the SD controller - could/should be done with GPIO functions
 *  Also configures some general purpose pins for MMC/SD functions
 */

#define WMT_GPIO_BASE	0xD8110000
#define GPIO_SD_ENABLE	0x40

void wmt_init_hardware(void)
{
	void __iomem *gpio_base;
	u32 reg_tmp;

	gpio_base = ioremap(WMT_GPIO_BASE, 65536);

	reg_tmp = readl(gpio_base + 0x68);
	writel(reg_tmp | 0x80, gpio_base + 0x68);
	reg_tmp = readl(gpio_base + 0x90);
	writel(reg_tmp | 0x80, gpio_base + 0x90);
	reg_tmp = readl(gpio_base + 0xB8);
	writel(reg_tmp & 0x7F, gpio_base + 0xB8);

	reg_tmp = readl(gpio_base + GPIO_SD_ENABLE);
	writel(reg_tmp & 0xFFFFFF00, gpio_base + GPIO_SD_ENABLE);

	iounmap(gpio_base);
}
/* End hacky code */


/* 
  Another hacky bit of code to program the PLL multipler - this should be done via other calls
  but PLL code isn't done yet
*/

#define WMT_PMC_BASE		0xD8130000
#define PMC_PLLB_MULTIPLIER	0x204
#define PMC_ENABLE_SD_CLK	0x254
#define PMC_SD_CLK		0x328

/* bit fields */
#define PMC_SD_CLK_EN		0x00040000

void wmt_setclock(struct mmc_host *mmc, u32 clock)
{
	struct wmt_mci_priv *priv;
	void __iomem *pmc_base;
	u32 reg_tmp;

	priv = mmc_priv(mmc);

	pmc_base = ioremap(WMT_PMC_BASE, 65536);

	/* 
	 * disable high-speed timing 
	 * high-speed timing required for >25Mhz - not supported at present
	 */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_EXTCTRL);
	writeb(reg_tmp & 0x7F, priv->sdmmc_base + SDMMC_EXTCTRL);

	reg_tmp = readl(pmc_base + PMC_PLLB_MULTIPLIER);

	if (clock <= 400000) {
		/* force sdmmc clkdiv to 64 */
		writel(reg_tmp | 0x20, pmc_base + PMC_SD_CLK);
	} else
	if (clock >= 25000000) {
		/* sdmmc clkdiv = pll multiplier */
		writel(reg_tmp, pmc_base + PMC_SD_CLK);
	} else {
		if ((reg_tmp * 25) % 20) {
			writel((reg_tmp * 25) / 20 + 1, pmc_base + PMC_SD_CLK);
		} else {
			writel((reg_tmp * 25) / 20, pmc_base + PMC_SD_CLK);
		}
	}

	/* enable clock */
	reg_tmp = readl(pmc_base + PMC_ENABLE_SD_CLK);
	writel(reg_tmp | PMC_SD_CLK_EN, pmc_base + PMC_ENABLE_SD_CLK);
	
	iounmap(pmc_base);
}

/* End hacky code */

void wmt_reset_hardware(struct mmc_host *mmc)
{
	struct wmt_mci_priv *priv;
	u32 reg_tmp;

	priv = mmc_priv(mmc);

	/* reset controller */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
	writeb(reg_tmp | BM_SOFT_RESET, priv->sdmmc_base + SDMMC_BUSMODE);

	/* reset response FIFO */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_CTLR);
	writeb(reg_tmp | CTLR_FIFO_RESET, priv->sdmmc_base + SDMMC_CTLR);

	/* enable GPI pin to detect card - necessary??? */
	writew(BLKL_INT_ENABLE | BLKL_GPI_CD, priv->sdmmc_base + SDMMC_BLKLEN);

	/* clear interrupt status */
	writeb(0xF1, priv->sdmmc_base + SDMMC_STS0);
	writeb(0xFE, priv->sdmmc_base + SDMMC_STS1);

	/* setup interrupts */
	writeb(INT0_CD_INT_EN | INT0_DI_INT_EN, priv->sdmmc_base + SDMMC_INTMASK0);
	writeb(0x00, priv->sdmmc_base + SDMMC_INTMASK1);

	/* set the DMA timeout */
	writew(8192, priv->sdmmc_base + SDMMC_DMATIMEOUT);

	/* auto clock freezing enable */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_STS2);
	writeb(reg_tmp | STS2_DIS_FORCECLK, priv->sdmmc_base + SDMMC_STS2);
	
	/* set a default clock speed of 400Khz */
	wmt_setclock(mmc, 400000);
}

void wmt_mci_start_command(struct wmt_mci_priv *priv)
{
	u32 reg_tmp;

	/* start command */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_CTLR);
	writeb(reg_tmp | CTLR_CMD_START, priv->sdmmc_base + SDMMC_CTLR);
}

int wmt_mci_send_command(struct mmc_host *mmc, u8 command, u8 cmdtype, u32 arg, u8 rsptype)
{
	struct wmt_mci_priv *priv;
	u32 reg_tmp;

	priv = mmc_priv(mmc);

	/* reset response FIFO */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_CTLR);
	writeb(reg_tmp | CTLR_FIFO_RESET, priv->sdmmc_base + SDMMC_CTLR);

	/* write command, arg, resptype registers */
	writeb(command, priv->sdmmc_base + SDMMC_CMD);
	writel(arg, priv->sdmmc_base + SDMMC_ARG);
	writeb(rsptype, priv->sdmmc_base + SDMMC_RSPTYPE);

	/* clear status bits */
	writeb(0xFF, priv->sdmmc_base + SDMMC_STS0);
	writeb(0xFF, priv->sdmmc_base + SDMMC_STS1);
	writeb(0xFF, priv->sdmmc_base + SDMMC_STS2);
	writeb(0xFF, priv->sdmmc_base + SDMMC_STS3);

	/* ensure SD on */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
	writeb(reg_tmp & (~BM_SD_OFF), priv->sdmmc_base + SDMMC_BUSMODE);

	/* set command type */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_CTLR);
	writeb((reg_tmp & 0x0F) | (cmdtype << 4), priv->sdmmc_base + SDMMC_CTLR);

	return 0;
}

int wmt_mci_post_command(struct mmc_host *mmc)
{
	struct wmt_mci_priv *priv;
	int idx1, idx2;
	u8 tmp_resp;
	u32 response;

	u32 status0;
	u32 status1;
	u32 status2;

	priv = mmc_priv(mmc);

	status0 = readb(priv->sdmmc_base + SDMMC_STS0);
	status1 = readb(priv->sdmmc_base + SDMMC_STS1);
	status2 = readb(priv->sdmmc_base + SDMMC_STS2);

	/*
	 *  wait for command to send
	 *  TODO: remove this code once interrupt handling is implemented
	 */
	while ((readb(priv->sdmmc_base + SDMMC_CTLR) & CTLR_CMD_START) != 0);

	/* wait for command to complete */
	while ((readb(priv->sdmmc_base + SDMMC_STS1) & STS1_RSP_TIMEOUT) != STS1_RSP_TIMEOUT) {
		if (( readb(priv->sdmmc_base + SDMMC_STS1) & STS1_CMDRSP_DONE) == STS1_CMDRSP_DONE)
			break;
	}

	/* check if command timed out */
	if ((readb(priv->sdmmc_base + SDMMC_STS1) & STS1_RSP_TIMEOUT) != 0) {
		return -ETIMEDOUT;
	}

	/* command executed ok - get the response */
	
	for (idx1=0; idx1<4; idx1++) {
		response = 0;
		for (idx2=0; idx2<4; idx2++) {
			if ((idx1 == 3) && (idx2 == 3))
				tmp_resp = readb(priv->sdmmc_base + SDMMC_RSP);
			else
				tmp_resp = readb(priv->sdmmc_base + SDMMC_RSP + (idx1*4) + idx2 + 1);
			response |= (tmp_resp << (idx2 * 8));
		}
		priv->req->cmd->resp[idx1] = cpu_to_be32(response);
	}

	writeb(status0, priv->sdmmc_base + SDMMC_STS0);
	writeb(status1, priv->sdmmc_base + SDMMC_STS1);
	writeb(status2, priv->sdmmc_base + SDMMC_STS2);

	return 0;
}

int wmt_dma_init(struct mmc_host *mmc)
{
	struct wmt_mci_priv *priv;
	
	priv = mmc_priv(mmc);

	writel(DMA_GCR_SOFT_RESET, priv->sdmmc_base + SDDMA_GCR);
	writel(DMA_GCR_DMA_EN, priv->sdmmc_base + SDDMA_GCR);
	if ((readl(priv->sdmmc_base + SDDMA_GCR) & DMA_GCR_DMA_EN) != 0)
		return 0;
	else
		return 1;
}

void wmt_dma_init_descriptor(struct wmt_dma_descriptor *desc, u16 req_count, u32 buffer_addr, u32 branch_addr, int end)
{	
	desc->flags = 0x40000000 | req_count;
	if (end)
		desc->flags |= 0x80000000;
	desc->data_buffer_addr = buffer_addr;
	desc->branch_addr = branch_addr;
}

#define PDMA_READ	0x00
#define PDMA_WRITE	0x01

void wmt_dma_config(struct mmc_host *mmc, u32 descaddr, u8 dir)
{
	struct wmt_mci_priv *priv;
	u32 reg_tmp;

	priv = mmc_priv(mmc);

	/* Enable DMA Interrupts */
	writel(DMA_IER_INT_EN, priv->sdmmc_base + SDDMA_IER);

	/* Write DMA Descriptor Pointer Register */
	writel(descaddr, priv->sdmmc_base + SDDMA_DESPR);

	writel(0x00, priv->sdmmc_base + SDDMA_CCR);

	if (dir == PDMA_WRITE) {
		reg_tmp = readl(priv->sdmmc_base + SDDMA_CCR);
		writel(reg_tmp & DMA_CCR_IF_TO_PERIPHERAL, priv->sdmmc_base + SDDMA_CCR);
	} else {
		reg_tmp = readl(priv->sdmmc_base + SDDMA_CCR);
		writel(reg_tmp | DMA_CCR_PERIPHERAL_TO_IF, priv->sdmmc_base + SDDMA_CCR);
	}
}

void wmt_dma_start(struct wmt_mci_priv *priv)
{
	u32 reg_tmp;

	reg_tmp = readl(priv->sdmmc_base + SDDMA_CCR);
	writel(reg_tmp | DMA_CCR_RUN, priv->sdmmc_base + SDDMA_CCR);
}

static void wmt_mci_request(struct mmc_host *mmc, struct mmc_request *req)
{
	struct wmt_mci_priv *priv;
	struct wmt_dma_descriptor *desc;
	struct completion datacomp;
	u8 command;
	u8 cmdtype;
	u32 arg;
	u8 rsptype;
	u32 reg_tmp;

	struct scatterlist *sg;
	int i;
	int sg_cnt;
	int offset;
	u32 dma_address;
	int desc_cnt;

	priv = mmc_priv(mmc);

	priv->req = req;

	command = req->cmd->opcode;
	arg = req->cmd->arg;
	rsptype = mmc_resp_type(req->cmd);
	cmdtype = 0;

	/* rsptype=7 only valid for SPI commands - should be =2 for SD */
	if (rsptype == 7) rsptype = 2;	
	/* rsptype=21 is R1B, convert for controller */
	if (rsptype == 21) rsptype = 9; 

	if (!req->data) {
		wmt_mci_send_command(mmc, command, cmdtype, arg, rsptype);
		wmt_mci_start_command(priv);
		req->cmd->error = wmt_mci_post_command(mmc);
		mmc_request_done(mmc, req);
	}
	if (req->data) {
		if (req->data->flags & MMC_DATA_WRITE) {
			if (req->data->blocks == 1)
				cmdtype = 1;
			else
				cmdtype = 3;

			wmt_dma_init(mmc);

			/* set controller data length */
			reg_tmp = readw(priv->sdmmc_base + SDMMC_BLKLEN);
			writew( (reg_tmp & 0xF800) | (req->data->blksz - 1), priv->sdmmc_base + SDMMC_BLKLEN);

			/* set controller block count (always 1 at the moment) */
			writew(req->data->blocks, priv->sdmmc_base + SDMMC_BLKCNT);

			desc = (struct wmt_dma_descriptor *)priv->dma_desc_buffer;
			
			sg_cnt = dma_map_sg(mmc_dev(mmc), req->data->sg, req->data->sg_len, DMA_TO_DEVICE);

			dma_address = priv->dma_desc_device_addr + 16;
			desc_cnt = 0;

			for_each_sg(req->data->sg, sg, sg_cnt, i) {
				offset = 0;
//				while ( offset < sg_dma_len(req->data->sg) ) {
				while ( offset < sg_dma_len(sg) ) {
//					wmt_dma_init_descriptor(desc, req->data->blksz, sg_dma_address(req->data->sg)+offset, dma_address, 0);
					wmt_dma_init_descriptor(desc, req->data->blksz, sg_dma_address(sg)+offset, dma_address, 0);
					desc++;
					desc_cnt++;
					offset += req->data->blksz;
					dma_address += 16;					
					if (desc_cnt == req->data->blocks) break;
				}				
			}
			desc--;
			desc->flags |= 0x80000000;

			wmt_dma_config(mmc, priv->dma_desc_device_addr, PDMA_WRITE);

			wmt_mci_send_command(mmc, command, cmdtype, arg, rsptype);

			priv->comp_dma = &datacomp;
			init_completion(priv->comp_dma);

			wmt_dma_start(priv);
			wmt_mci_start_command(priv);

			wait_for_completion(priv->comp_dma);

			req->data->bytes_xfered = req->data->blksz * req->data->blocks;

			/* unmap the DMA pages used for write data */
			dma_unmap_sg(mmc_dev(mmc), req->data->sg, req->data->sg_len, DMA_TO_DEVICE);

			/* Check if the DMA ISR returned a data error */
			if (req->data->error != 0) {
				req->cmd->error = req->data->error;
			} else {
				req->cmd->error = wmt_mci_post_command(mmc);
				req->data->error = req->cmd->error;

				/* TODO: Check this - data->stop command may be related */
				/* if >1 block requested, send a stop command to end transmission */
				if (req->data->blocks > 1) {
					wmt_mci_send_command(mmc, MMC_STOP_TRANSMISSION, 7, arg, 9);
					wmt_mci_start_command(priv);
					req->cmd->error = wmt_mci_post_command(mmc);
					req->data->error = req->cmd->error;
				}
			}
			
		} else {
			/* execute a read command */
			if (req->data->blocks == 1)
				cmdtype = 2;
			else
				cmdtype = 4;

			wmt_dma_init(mmc);

			/* set controller data length */
			reg_tmp = readw(priv->sdmmc_base + SDMMC_BLKLEN);
			writew( (reg_tmp & 0xF800) | (req->data->blksz - 1), priv->sdmmc_base + SDMMC_BLKLEN);

			/* set controller block count */
			writew(req->data->blocks, priv->sdmmc_base + SDMMC_BLKCNT);

			desc = (struct wmt_dma_descriptor *)priv->dma_desc_buffer;
			
			sg_cnt = dma_map_sg(mmc_dev(mmc), req->data->sg, req->data->sg_len, DMA_FROM_DEVICE);

			dma_address = priv->dma_desc_device_addr + 16;
			desc_cnt = 0;

			for_each_sg(req->data->sg, sg, sg_cnt, i) {
				offset = 0;
				while ( offset < sg_dma_len(sg) ) {
//				while ( offset < sg_dma_len(req->data->sg) ) {
//					wmt_dma_init_descriptor(desc, req->data->blksz, sg_dma_address(req->data->sg)+offset, dma_address, 0);
					wmt_dma_init_descriptor(desc, req->data->blksz, sg_dma_address(sg)+offset, dma_address, 0);
					desc++;
					desc_cnt++;
					offset += req->data->blksz;
					dma_address += 16;
					if (desc_cnt == req->data->blocks) break;
				}
			}
			
			desc--;
			desc->flags |= 0x80000000;

			wmt_dma_config(mmc, priv->dma_desc_device_addr, PDMA_READ);

			wmt_mci_send_command(mmc, command, cmdtype, arg, rsptype);

			priv->comp_dma = &datacomp;
			init_completion(priv->comp_dma);

			wmt_dma_start(priv);
			wmt_mci_start_command(priv);

			wait_for_completion(priv->comp_dma);

			priv->req->data->bytes_xfered = (req->data->blksz * desc_cnt);
			dma_unmap_sg(mmc_dev(mmc), req->data->sg, req->data->sg_len, DMA_FROM_DEVICE);

			/* Check if the DMA ISR returned a data error */
			if (req->data->error != 0) {
				req->cmd->error = req->data->error;
			} else {
				req->cmd->error = wmt_mci_post_command(mmc);
				req->data->error = req->cmd->error;

				/* TODO: Check this - data->stop command may be related */
				/* if >1 block requested, send a stop command to end transmission */
				if (req->data->blocks > 1) {
					wmt_mci_send_command(mmc, MMC_STOP_TRANSMISSION, 7, arg, 9);
					wmt_mci_start_command(priv);
					req->cmd->error = wmt_mci_post_command(mmc);
					req->data->error = req->cmd->error;
				}
			}
		}
		mmc_request_done(mmc, req);
	}
}

static void wmt_mci_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	struct wmt_mci_priv *priv;
	u32 reg_tmp;

	priv = mmc_priv(mmc);

	if (ios->power_mode == MMC_POWER_UP) {
		wmt_reset_hardware(mmc);

		reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
		writeb(reg_tmp & (~BM_SD_OFF), priv->sdmmc_base + SDMMC_BUSMODE);
	}
	if (ios->power_mode == MMC_POWER_OFF) {
		reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
		writeb(reg_tmp | BM_SD_OFF, priv->sdmmc_base + SDMMC_BUSMODE);
	}

	/* 
	 * setup the requested clock speed
	 */
	wmt_setclock(mmc, ios->clock);

	switch (ios->bus_width) {
		case MMC_BUS_WIDTH_8:
			/* 
			 * This shouldn't get called as SDv2 only supports up to 1 or 4 bit mode 
			 * Controller doesnt support SDv3
			 * Also required for MMC cards - not implemented yet
			 */
			reg_tmp = readb(priv->sdmmc_base + SDMMC_EXTCTRL);
			writeb(reg_tmp | 0x04, priv->sdmmc_base + SDMMC_EXTCTRL);
			break;
		case MMC_BUS_WIDTH_4:
			reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
			writeb(reg_tmp | BM_FOURBIT_MODE, priv->sdmmc_base + SDMMC_BUSMODE);

			reg_tmp = readb(priv->sdmmc_base + SDMMC_EXTCTRL);
			writeb(reg_tmp & 0xFB, priv->sdmmc_base + SDMMC_EXTCTRL);
			break;
		case MMC_BUS_WIDTH_1:
			reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
			writeb(reg_tmp & BM_ONEBIT_MASK, priv->sdmmc_base + SDMMC_BUSMODE);

			reg_tmp = readb(priv->sdmmc_base + SDMMC_EXTCTRL);
			writeb(reg_tmp & 0xFB, priv->sdmmc_base + SDMMC_EXTCTRL);
			break;
	}
}

static int wmt_mci_get_ro(struct mmc_host *mmc)
{
	struct wmt_mci_priv *priv;
	
	priv = mmc_priv(mmc);
	return (((readb(priv->sdmmc_base + SDMMC_STS0) >> 1) ^ 1) & 1);
}

static struct mmc_host_ops wmt_mci_ops = {
	.request = wmt_mci_request,
	.set_ios = wmt_mci_set_ios,
	.get_ro = wmt_mci_get_ro,
};

static int __init wmt_mci_probe(struct platform_device *pdev)
{
	struct mmc_host *mmc;
	struct wmt_mci_priv *priv;
	struct resource *res;
	int ret;
	int platform_irq1;
	int platform_irq2;
	
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		printk(KERN_ERR "[MMC/SD] Getting platform resource failed: Memory-mapped IO\n");
		return -ENXIO;
	}

	if (!request_mem_region(res->start, res->end - res->start + 1, DRIVER_NAME)) {
		printk(KERN_ERR "[MMC/SD] Getting memory region failed\n");
		return -EBUSY;
	}

	/*
	 * Two irqs should be defined, first for regular ISR, second for DMA ISR
	 */
	platform_irq1 = platform_get_irq(pdev, 0);
	platform_irq2 = platform_get_irq(pdev, 1);

	if ((platform_irq1 == -1) | (platform_irq2 == -1)) {
		printk(KERN_ERR "[MMC/SD] Getting platform IRQs failed!\n");
		ret = -ENXIO;
		goto fail1;
	}

	mmc = mmc_alloc_host(sizeof(struct wmt_mci_priv), &pdev->dev);
	if (!mmc) {
		printk(KERN_ERR "[MMC/SD] Failed to allocate mmc_host\n");
		ret = -ENOMEM;
		goto fail1;
	}

	mmc->ops = &wmt_mci_ops;
	mmc->f_min = 390425;
	mmc->f_max = 50000000;
	mmc->ocr_avail = MMC_VDD_32_33 | MMC_VDD_33_34;
	mmc->caps = MMC_CAP_4_BIT_DATA | MMC_CAP_MMC_HIGHSPEED | MMC_CAP_SD_HIGHSPEED;
	mmc->max_seg_size = 65024;
	mmc->max_segs = 128;
	mmc->max_req_size = 1048576;
	mmc->max_blk_size = 2048;
	mmc->max_blk_count = 128;

	priv = mmc_priv(mmc);
	priv->mmc = mmc;

	priv->sdmmc_base = ioremap(res->start, res->end - res->start + 1);
	if (!priv->sdmmc_base) {
		printk(KERN_ERR "[MMC/SD] Failed to map IO space\n");
		ret = -ENOMEM;
		goto fail2;
	}

	priv->irq_regular = platform_irq1;
	priv->irq_dma = platform_irq2;

	ret = request_irq(platform_irq1, wmt_mci_regular_isr, 0, "sdmmc", priv);
	if (ret) {
		printk(KERN_ERR "[MMC/SD] Failed to register interrupt handler (regular)\n");
		goto fail3;
	}

	ret = request_irq(platform_irq2, wmt_mci_dma_isr, 32, "sdmmc", priv);
	if (ret) {
		printk(KERN_ERR "[MMC/SD] Failed to register interrupt handler (DMA)\n");
		goto fail4;
	}

	/* alloc some DMA buffers for descriptors/transfers */
	priv->dma_desc_buffer = dma_alloc_coherent(&pdev->dev, 2048, &priv->dma_desc_device_addr, 208);
	if (!priv->dma_desc_buffer) {
		printk(KERN_ERR "[MMC/SD] Failed to allocate DMA descriptor memory\n");
		ret = -EPERM;
		goto fail5;
	}

	platform_set_drvdata(pdev, mmc);

	/* initialize non-controller stuff that we rely on - Read notes @ function definition */
	wmt_init_hardware();

	/* configure the controller to a known 'ready' state */
	wmt_reset_hardware(mmc);

	mmc_add_host(mmc);

	printk(KERN_INFO "[MMC/SD] WMT MCI Controller registered\n");	

	return 0;
fail5:
	free_irq(platform_irq1, priv);
fail4:
	free_irq(platform_irq1, priv);
fail3:
	iounmap(priv->sdmmc_base);
fail2:
	mmc_free_host(mmc);
fail1:
	release_mem_region(res->start, res->end - res->start + 1);
	return ret;	
}

static int __exit wmt_mci_remove(struct platform_device *pdev)
{
	struct mmc_host *mmc;
	struct wmt_mci_priv *priv;
	struct resource *res;
	u32 reg_tmp;

	mmc = platform_get_drvdata(pdev);
	if (!mmc) {
		printk(KERN_DEBUG "[MMC/SD] Invalid driver data during device removal\n");
		return -1;
	}

	priv = mmc_priv(mmc);

	mmc_remove_host(mmc);

	/* reset SD controller */
	reg_tmp = readb(priv->sdmmc_base + SDMMC_BUSMODE);
	writel(reg_tmp | BM_SOFT_RESET, priv->sdmmc_base + SDMMC_BUSMODE);

	/* release the dma buffers */
	dma_free_coherent(&pdev->dev, 2048, priv->dma_desc_buffer, priv->dma_desc_device_addr);
	

	free_irq(priv->irq_regular, priv);
	free_irq(priv->irq_dma, priv);

	iounmap(priv->sdmmc_base);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(res->start, res->end - res->start + 1);

	mmc_free_host(mmc);

	platform_set_drvdata(pdev, NULL);

	printk(KERN_INFO "[MMC/SD] WMT MCI device removed\n");

	return 0;
}


static struct platform_driver wmt_mci_driver = {
	.remove = __exit_p(wmt_mci_remove),
	.suspend = NULL,
	.resume = NULL,
	.driver = {
		.name = DRIVER_NAME,
		.owner = THIS_MODULE,
	},
};

static int __init wmt_mci_init(void)
{
	return platform_driver_probe(&wmt_mci_driver, &wmt_mci_probe);
}

static void __exit wmt_mci_exit(void)
{
	platform_driver_unregister(&wmt_mci_driver);
}

module_init(wmt_mci_init);
module_exit(wmt_mci_exit);

MODULE_DESCRIPTION("WMT MMC/SD Driver");
MODULE_AUTHOR("Tony Prisk");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:wmt-sdmmc");
